// Pawe Kaczorowski, Album: 206096, Grupa: 5TIZ
// Temat: biblioteka stl : opis kontenera, bit_vector.

// Bit_vector jest to w zasadzie vector<bool> i posiada ten sam interface co vector. Rnica jest
// w efektywnoci pamiciowej. Oglnie vector zajmuj co najmniej 1 bit na elemnent, a bit_vector dokadnie 1 bit na element.
// Nazwa vector<bool> jest mylca poniewa nie s przechowywane wartoci typu bool, a wartoci bitowe.

// Nazwa bit_vector jest to wic typedef i nie jest to zawarte w standardzie C++.

// Przykadem zastosowania bdzie krtki program, ktry dodaje dwie liczby binarne. Bit_vector bdzie odzwierciedleniem
// liczby binarnej. Program posiada funkcj, ktra dodaje wektory w zaoeniu e s to liczby binarne
// Wynikiem dziaania jest przedstawienie dziaa na liczbach binarnych, jak i dziesitnych (po przeliczeniu).

// Doczamy <vector> aby korzysta z vector<bool>
#include <vector> 
#include <iostream> 
// Doczamy aby korzysta z generatora liczb losowych.
#include <boost/random.hpp>
// Doczamy math aby korzysta z potgowania.
#include <cmath>

using namespace std;

//Zakadamy sta dugo wektorw
const int SIZE = 16;

//Definiujemy typ vector<bool> aby dalej posugiwa si nazw bit_vector.
typedef vector<bool> bit_vector;

//Funkcja, ktra suy jako blok funkcji dodawania(wchodz dwa skadniki sumy i bit przeniesienia  z poprzedniego bloku
bool add(bool a, bool b, bool* ca_in, bool* ca_out);

//Funkcja dodajca dwa liczby binarne zapisane w bit_vector tej samej dugoci
bit_vector add_vector(bit_vector a, bit_vector b, bool* ca);

//Funkcja wypisujca bit_vector na standardowe wyjcie
void print_vec(bit_vector a, int size);

//Funckcja przeliczajca warto wektora binarnego na liczb dzisitn
int binary_to_dec(bit_vector a, int size);

int main () { 

	//Definiujemy bit przeniesienia
	bool carry = false;
	bool* ca = &carry;
	
	//Definiujemy przykadowe wektory
	bit_vector V(SIZE);
	bit_vector J(SIZE);
	bit_vector L(SIZE,false);

	//Zdefiniowanie generatora bitw ( zwraca warto 0 lub 1)
	boost::mt19937 rng;               
	boost::uniform_int<> bit(0,1);  
	boost::variate_generator<boost::mt19937&,boost::uniform_int<> > die(rng, bit);

	//Wypeniamy wektory wartociami losowymi, po czym drukujemy ich wartoci
	for(int i=0; i<SIZE ;i++){
		switch(die()){
			case 0: V[i] = false; break;
			case 1: V[i] = true;  break;
		}
	}
	
	cout<<"Losowo wypelniony pierwszy wektor "<<SIZE<<" bitowy"<<endl;
	print_vec(V, SIZE);
	
	for(int i=0; i<SIZE ;i++){
		switch(die()){
			case 0: J[i] = false; break;
			case 1: J[i] = true;  break;
		}
	}
	
	cout<<"Losowo wypelniony drugi wektor "<<SIZE<<" bitowy"<<endl;
	print_vec(J, SIZE);

	cout<<"Pusty wektor trzeci:"<<endl;
	print_vec(L,SIZE);

	// Dodajemy 2 pierwsze wektory, a wynik zapisujemy do pustego wektora.
	L = add_vector(V,J,ca);

	// Drukujemy na wyjcie.
	cout<<"Wynik dodania wektora pierwszego do wektora drugiego(zajmuje "<<SIZE<<" bity + bit przeniesienia)"<<endl;
	cout<<carry;
	print_vec(L,SIZE);

	//Konwersja na wartoci dziesitne.
	cout<<"Zapis dziesietny operacji:"<<endl;
	cout<<binary_to_dec(V,SIZE)<<" + "<<binary_to_dec(J,SIZE)<<" = "<<binary_to_dec(L,SIZE)<<endl;
	

  getchar();
  return 0; 
} 

// Funkcje

// Blok funkcyjny w dodawaniu binarnym.
bool add(bool a, bool b, bool* ca_in, bool* ca_out){
	if(a==false && b==false && *ca_in==false){ca_out = false;return false;}
	else if(a==false && b==false && *ca_in==true){*ca_out = false;return true;}
	else if(a==false && b==true && *ca_in==false){*ca_out = false;return true;}
	else if(a==false && b==true && *ca_in==true){*ca_out = true;return false;}
	else if(a==true && b==false && *ca_in==false){*ca_out = false;return true;}
	else if(a==true && b==false && *ca_in==true){*ca_out = true;return false;}
	else if(a==true && b==true && *ca_in==false){*ca_out = true;return false;}
	else {*ca_out = true;return true;}
}

// Dodawanie dwch  bit_vector
bit_vector add_vector(bit_vector a, bit_vector b, bool* ca){
	bit_vector V(a.size());
	for(int i = a.size()-1; i >= 0; i--){
		V[i] = add(a[i], b[i], ca, ca);
	}
return V;
}

// Drukowanie na wyjcie
void print_vec(bit_vector a, int size){
	for(int i=0; i<size;i++){
		cout<<a[i];
	}
	cout<<endl<<endl;
}

//Konwersja bin > dec
int binary_to_dec(bit_vector a, int size){
	double result = 0;
	for(int i=0; i<size;i++){
		result+=(a[i]*pow((double)2,(double)(size-1-i)));
	}
	return (int)result;
}
